package be.iminds.ilabt.jfed.experimenter_gui.config;

import be.iminds.ilabt.jfed.fedmon.webapi.service.json.*;
import be.iminds.ilabt.jfed.rspec.model.DiskImage;
import be.iminds.ilabt.jfed.rspec.model.ModelRspec;
import be.iminds.ilabt.jfed.rspec.model.NodeLocation;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import be.iminds.ilabt.jfed.rspec.rspec_source.AdvertisementRspecSource;
import be.iminds.ilabt.jfed.util.common.GeniUrn;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.ReadOnlyListProperty;
import javafx.beans.property.ReadOnlyObjectProperty;
import javafx.collections.ObservableList;
import javafx.scene.paint.Color;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.*;

/**
 * JFedGuiConfig
 *
 * replaces JFedConfiguration with config that comes from server instead of file
 */
public interface JFedGuiConfig {
    String FLAG_ESPEC = "espec";
    String FLAG_SHOW_TUTORIAL_BUTTON = "showTutorialButton";
    String FLAG_SLA_AVAILABLE = "slaAvailable";
    String FLAG_QOE_REPORTING = "reportQOE";
    String FLAG_SHOW_REPUTATION = "showReputation";
    String FLAG_DISABLE_EXP_DELETE = "disableExperimentDelete"; //prevent Delete/Terminate on experiment
    String FLAG_DISABLE_EXP_ACTION = "disableExperimentAction"; //prevent editing keys, sharing, restart, etc of experiment
    String FLAG_DISABLE_EXP_RENEW  = "disableExperimentRenew";  //prevent renew of experiment
    String FLAG_DISABLE_EXP_START  = "disableExperimentStart";  //prevent starting and experiment
    String FLAG_DISABLE_PREFERENCES = "disablePreferences";  //prevent starting and experiment

    /**
     * The fixed ID of the automatically generated ResourceClass that contains all servers that are not
     * referenced in any Resource.
     */
    String RESOURCE_CLASS_ID_OTHER_SERVERS = "other-servers";

    @Nullable GtsConfig getGtsConfig();

    boolean hasFlag(String flagName);
    @Nullable Boolean getBooleanFlag(String flagName);
    @Nullable String getStringFlag(String flagName);
    @Nullable Integer getIntegerFlag(String flagName);
    @Nullable Long getLongFlag(String flagName);

    default boolean isShowTutorialButtonEnabled() {
        return Objects.equals(Boolean.TRUE, getBooleanFlag(FLAG_SHOW_TUTORIAL_BUTTON));
    }
    default boolean isEspecEnabled() {
        return Objects.equals(Boolean.TRUE, getBooleanFlag(FLAG_ESPEC));
    }
    default boolean isSlaAvailable() {
        return Objects.equals(Boolean.TRUE, getBooleanFlag(FLAG_SLA_AVAILABLE));
    }
    default boolean isQOEReportingEnabled() {
        return Objects.equals(Boolean.TRUE, getBooleanFlag(FLAG_QOE_REPORTING));
    }
    default boolean isShowReputationEnabled() {
        return Objects.equals(Boolean.TRUE, getBooleanFlag(FLAG_SHOW_REPUTATION));
    }
    default boolean isExperimentDeleteDisabled() {
        return Objects.equals(Boolean.TRUE, getBooleanFlag(FLAG_DISABLE_EXP_DELETE));
    }
    default boolean isExperimentActionDisabled() {
        return Objects.equals(Boolean.TRUE, getBooleanFlag(FLAG_DISABLE_EXP_ACTION));
    }
    default boolean isExperimentRenewDisabled() {
        return Objects.equals(Boolean.TRUE, getBooleanFlag(FLAG_DISABLE_EXP_RENEW));
    }
    default boolean isExperimentStartDisabled() {
        return Objects.equals(Boolean.TRUE, getBooleanFlag(FLAG_DISABLE_EXP_START));
    }
    default boolean isPreferencesDisabled() {
        return Objects.equals(Boolean.TRUE, getBooleanFlag(FLAG_DISABLE_PREFERENCES));
    }

    /**
     * guarantees:
     * The ResourceClass point to actual Resources with getResources and getDefaultResource
     * These Resources point to the same Server objects that TestbedInfoSource points to.
     * (So all server info is filled int, and they point to complete testbed and services objects)
     * */
    @Nullable ResourceClass getResourceClass(String id);
    /**
     * guarantees:
     * The ResourceClass point to actual Resources with getResources and getDefaultResource
     * These Resources point to the same Server objects that TestbedInfoSource points to.
     * (So all server info is filled int, and they point to complete testbed and services objects)
     * */
    @Nonnull Collection<ResourceClass> getResourceClasses();

    /** return the list of Triggered messages. */
    @Nonnull List<TriggeredMessage> getTriggeredMessages();


    enum ResourceType {
        RAW_PC("Bare Metal"),
        VM("VMs"),
        OTHER("Other");

        ResourceType(String friendlyName) {
            this.friendlyName = friendlyName;
        }

        private final String friendlyName;

        public String getFriendlyName() {
            return friendlyName;
        }
    };
    /*
     * detection of specific resource types contained in a resourceClass
     * if the resourceClass is not a "leaf", this is the type of the default.
     *
     * (don't figure this out manually using the ResourceClass etc, let this class determine it.)
     * */
    @Nonnull ResourceType getResourceType(@Nonnull ResourceClass resourceClass);

    /** @return never null, but can return empty list. */
    @Nonnull Collection<ResourceClass> getResourceClassesByType(@Nonnull ResourceType resourceType);

    @Nullable ResourceClass getDefaultResourceClass();



    /**
     * exactly the same Server objects that TestbedInfoSource has.
     * (So all server info is filled int, and they point to complete testbed and services objects)
     * */
    @Nonnull List<Server> getBindableServers();
    @Nonnull List<Server> getAllServersInVisibleResources();
    @Nonnull List<Server> getHiddenServers();


    @Nullable ResourceClass findBestResourceClass(@Nonnull RspecNode node);
    @Nonnull List<ResourceClass> findAllPossibleResourceClasses(@Nonnull RspecNode node);

    @Nonnull List<Resource> findAllPossibleResources(@Nonnull Server server);
    @Nonnull List<ResourceClass> findAllPossibleResourceClasses(@Nonnull Server server);

    @Nullable Resource findBestResource(@Nullable ResourceClass resourceClass, @Nonnull RspecNode node, @Nullable Server overrideComponentManager);

    @Nonnull List<Resource> findAllPossibleResources(@Nullable ResourceClass resourceClass, @Nonnull RspecNode node, @Nullable Server overrideComponentManager);

    //same but with null parameters dropped
    @Nullable default Resource findBestResource(@Nullable ResourceClass resourceClass, RspecNode node) {
        return findBestResource(resourceClass, node, null);
    }
    @Nonnull default List<Resource> findAllPossibleResource(@Nullable ResourceClass resourceClass, RspecNode node) {
        return findAllPossibleResources(resourceClass, node, null);
    }
    @Nullable default Resource findBestResource(RspecNode node, @Nullable Server overrideComponentManager) {
        return findBestResource(null, node, overrideComponentManager);
    }
    @Nonnull default List<Resource> findAllPossibleResource(RspecNode node, @Nullable Server overrideComponentManager) {
        return findAllPossibleResources(null, node, overrideComponentManager);
    }
    @Nullable default Resource findBestResource(RspecNode node) {
        return findBestResource(null, node, null);
    }
    @Nonnull default List<Resource> findAllPossibleResource(RspecNode node) {
        return findAllPossibleResources(null, node, null);
    }

    /**
     * pick a random default Resource for the resourceClass.
     * For each jFed session the same is always chosen.
     * */
    @Nullable Resource getDefaultResource(ResourceClass resourceClass);


    //TestbedAdvertisementCache functionality:
    /**
     * Find out if a node for this sliver type should be exclusive or not.
     * Ideally this info should come from a cache of advertisement RSpecs.
     * */
    @Nullable Boolean getNodeExclusivity(GeniUrn componentManagerUrn, String sliverTypeName);

    /*
       callers don't need to care where this info comes from.
       Note: These are JavaFX ListProperty, so you can listen to changes.
       These are ReadOnlyListProperty, so you can't change them using these accessors.
     */
    enum Scope {
        ALL,
        AVAILABLE_ONLY;

        public static Scope getScopeForAvailable(boolean available) {
            return available ? AVAILABLE_ONLY : ALL;
        }
        public static Scope getScopeForAll(boolean all) {
            return !all ? AVAILABLE_ONLY : ALL;
        }
    };
    boolean hasUpToDateComponentInfo(@Nonnull Server server, @Nonnull Scope scope);
    ObservableList<JFedGuiConfigImpl.ComponentInfo> getComponentInfo(@Nonnull Server server, @Nonnull Scope scope, @Nullable Resource resource);
    ObservableList<String> getSliverTypeNames(@Nonnull Server server);
    ObservableList<HardwareTypeInfo> getHardwareTypes(@Nonnull Server server, @Nonnull Scope scope, @Nullable Resource resource);
    ReadOnlyListProperty<DiskImage> getDiskImages(@Nonnull Server server, @Nullable String sliverTypeName);
    default ReadOnlyListProperty<DiskImage> getDiskImages(@Nonnull Server server) {
        return getDiskImages(server, null);
    }
    default ReadOnlyListProperty<DiskImage> getDiskImages(@Nonnull Resource resource) {
        return getDiskImages(resource.getServer(), resource.getSliverType());
    }

    @Nonnull ObjectProperty<Integer> getServerReputation(@Nonnull Server server);
    @Nonnull ObservableList<MaintenanceInfo> getServerMaintenance(@Nonnull Server server);
    @Nonnull ObjectProperty<ServerGlimpse.HealthInfo> getServerHealthInfo(@Nonnull Server server);
    @Nonnull ObjectProperty<ServerGlimpse.AdvertisementInfo.HardwareTypeInfo> getHardwareTypeInfo(@Nonnull Server server);

    @Nonnull
    ReadOnlyObjectProperty<? extends AdvertisementCounts>
        getServerNodeCounts(@Nonnull Server server);

    //TODO: register new advertisement info?  very probably does not need to be in this interface:
    //        AdvertisementFetchManager should make sure the actual implementation gets the info somehow

    /** Get advertisement. Note: it is preferable to get the info from it using the getters above, but this is available for cases that is not enough.
     * use AdvertisementFetchManager to request advertisement refresh/initial fetch
     * you can add listeners to this property, it will get updated.
     *
     * There is no need tpo manually set this ever. The AdvertisementFetchManager is responsible for this,
     * and will do this BEFORE completing it's fetch task and calling any listeners of that Task.
     *
     * @param availableOnly if set, specifies which advertisement (available or all) is needed. If not set, any will be returned,
     *                      with a preference for the "all" advertisement.
     *
     * @return never null, but the value of the property can be null
     */
    ReadOnlyObjectProperty<AdvertisementRspecSource> getAdvertisement(Server server, @Nullable Boolean availableOnly);


    //NodesVisualization functionality:
    //TODO fedmon-v3 refactor: for now this will work using the old system underneath
    enum NodesVisualization {
        MAP,  /* nodes are visualized on a world map */
        IMAGE /* nodes are visualized on an image, on which they all have defined coordinates */
    }
    NodesVisualization getNodesVisualization(Server server);
    ImageMap<GeniUrn> getImageMap(Server server);


    //HealthStatus functionality
    @Nullable ServerGlimpse getServerGlimpse(Server server);


    void refreshServerGlimpsesIfNeeded();
    void forceRefreshServerGlimpses();


    //various helpers/converters
    @Nonnull Color getColor(Server server);
    @Nonnull Color getColor(Testbed testbed);
    @Nonnull Color getColor(String colorName);

    public static final Comparator<ComponentInfo> COMPONENT_INFO_BY_ID_COMPARATOR = new Comparator<ComponentInfo>() {
        @Override
        public int compare(ComponentInfo o1, ComponentInfo o2) {
            String s1 = o1.getName() == null ? o1.getUrn().toString() : o1.getName();
            String s2 = o2.getName() == null ? o2.getUrn().toString() : o2.getName();

            if (s1 == null || s2 == null) {
                throw new NullPointerException("ComponentInfo must always have some ID");
            }
            return s1.compareTo(s2);
        }
    };

    public static class HardwareTypeInfo implements Comparable {
        private final String hardwareTypeName;
        private final Boolean available;

        public HardwareTypeInfo(String hardwareTypeName) {
            this(hardwareTypeName, null);
        }

        public HardwareTypeInfo(String hardwareTypeName, Boolean available) {
            assert hardwareTypeName != null;
            this.hardwareTypeName = hardwareTypeName;
            this.available = available;
        }

        public String getHardwareTypeName() {
            return hardwareTypeName;
        }

        public Boolean isAvailable() {
            return available;
        }

        @Override
        public int compareTo(@Nonnull Object o) {
            return hardwareTypeName.compareTo(((HardwareTypeInfo) o).hardwareTypeName);
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            HardwareTypeInfo that = (HardwareTypeInfo) o;

            if (!hardwareTypeName.equals(that.hardwareTypeName)) return false;

            return true;
        }

        @Override
        public int hashCode() {
            return hardwareTypeName.hashCode();
        }

        @Override
        public String toString() {
            return hardwareTypeName;
        }
    }

    public static class ComponentInfo implements Comparable {
        private static final Logger LOG = LoggerFactory.getLogger(ComponentInfo.class);

        @Nullable private final String name;
        @Nullable private final GeniUrn urn;
        @Nullable private final NodeLocation nodeLocation;
        private final boolean available;

        public static enum ExtraStatusInfoType { NODE_LOGIN, SENSOR };
        //Extra info about (partial) status of node, such as "did last nod login test succeed?" and "did last sensor check succeed?"
        @Nonnull private final Map<ExtraStatusInfoType, Boolean> extraStatusInfo;
        //Extra info extracted while testing the node. This can include wireless sensor info, BIOS version, etc...
        @Nonnull private final Map<String, String> extraInfo;

        @Nullable private final List<String> hardwareTypeNames;
        @Nullable private final List<String> sliverTypeNames;

        public ComponentInfo(@Nullable String name,
                             @Nullable GeniUrn urn,
                             boolean available,
                             @Nullable Map<ExtraStatusInfoType, Boolean> extraStatusInfo,
                             @Nullable Map<String, String> extraInfo,
                             @Nullable NodeLocation nodeLocation,
                             @Nullable Collection<String> hardwareTypeNames, @Nullable Collection<String> sliverTypeNames) {
            //we need at least one identifier
            assert name != null || urn != null;

            this.name = name;
            this.urn = urn;
            this.available = available;
            this.nodeLocation = nodeLocation;
            this.hardwareTypeNames = hardwareTypeNames == null ? null : new ArrayList<>(hardwareTypeNames);
            this.sliverTypeNames = sliverTypeNames == null ? null : new ArrayList<>(sliverTypeNames);

            if (extraStatusInfo == null) {
                this.extraStatusInfo = Collections.emptyMap();
            } else {
                EnumMap<ExtraStatusInfoType, Boolean> newExtraStatusInfo =
                        new EnumMap<ExtraStatusInfoType, Boolean>(ExtraStatusInfoType.class);
                newExtraStatusInfo.putAll(extraStatusInfo);
                this.extraStatusInfo = Collections.unmodifiableMap(newExtraStatusInfo);
            }

            if (extraInfo == null) {
                this.extraInfo = Collections.emptyMap();
            } else {
                this.extraInfo = Collections.unmodifiableMap(new HashMap<>(extraInfo));
            }
        }

        @Nonnull
        public static ComponentInfo create(@Nonnull GeniUrn urn,
                             boolean available,
                             @Nonnull ExtraStatusInfoType extraStatusInfoType, boolean status,
                             @Nullable Map<String, String> extraInfo) {
            return new ComponentInfo(null, urn, available,
                    Collections.singletonMap(extraStatusInfoType, status),
                    extraInfo,
                    null, null, null);
        }


        @Nonnull
        public ComponentInfo copyWithAdditionalStatusInfo(@Nonnull ExtraStatusInfoType extraStatusInfoType, boolean status) {
            EnumMap<ExtraStatusInfoType, Boolean> newExtraStatusInfo =
                    new EnumMap<ExtraStatusInfoType, Boolean>(ExtraStatusInfoType.class);
            newExtraStatusInfo.putAll(this.extraStatusInfo);
            newExtraStatusInfo.put(extraStatusInfoType, status);

            return new ComponentInfo(
                    this.getName(),
                    this.getUrn(),
                    this.available,
                    newExtraStatusInfo,
                    this.extraInfo,
                    this.nodeLocation,
                    this.hardwareTypeNames,
                    this.sliverTypeNames
            );
        }
        @Nonnull
        public ComponentInfo copyWithAdditionalExtraInfo(@Nullable Map<String, String> additionalExtraInfo) {
            Map<String, String> newExtraInfo;
            if (additionalExtraInfo == null) {
                newExtraInfo = Collections.emptyMap();
            } else {
                newExtraInfo = new HashMap<>(this.extraInfo);
                newExtraInfo.putAll(additionalExtraInfo);
            }

            return new ComponentInfo(
                    this.getName(),
                    this.getUrn(),
                    this.available,
                    this.extraStatusInfo,
                    newExtraInfo,
                    this.nodeLocation,
                    this.hardwareTypeNames,
                    this.sliverTypeNames
            );
        }

        @Nonnull
        public ComponentInfo copyWithMergedExtraInfo(@Nonnull ComponentInfo other) {
            EnumMap<ExtraStatusInfoType, Boolean> newExtraStatusInfo =
                    new EnumMap<ExtraStatusInfoType, Boolean>(ExtraStatusInfoType.class);
            newExtraStatusInfo.putAll(this.extraStatusInfo);
            newExtraStatusInfo.putAll(other.extraStatusInfo);

            Map<String, String> newExtraInfo = new HashMap<>(this.extraInfo);
            newExtraInfo.putAll(other.extraInfo);

            return new ComponentInfo(
                    this.getName(),
                    this.getUrn(),
                    this.available,
                    newExtraStatusInfo,
                    newExtraInfo,
                    this.nodeLocation,
                    this.hardwareTypeNames,
                    this.sliverTypeNames
            );
        }
        @Nonnull
        public ComponentInfo copyWithAdditionalExtraInfo(@Nonnull String key, @Nonnull String value) {
            return copyWithAdditionalExtraInfo(Collections.singletonMap(key, value));
        }

        @Nullable
        public String getName() {
            return name;
        }

        @Nullable
        public GeniUrn getUrn() {
            return urn;
        }

        public boolean isAvailableInRspec() {
            return available;
        }

        public boolean isAvailable() {
            if (Objects.equals(isNodeLoginTestSuccess(), Boolean.FALSE))
                return false;
            return available;
        }

        public boolean hasExtraStatus(@Nonnull ExtraStatusInfoType type) {
            return extraStatusInfo.containsKey(type);
        }

        @Nullable
        public Boolean isSuccess(@Nonnull ExtraStatusInfoType type) {
            if (extraStatusInfo.containsKey(type))
                return extraStatusInfo.get(type);
            else
                return null;
        }
        @Nullable
        public Boolean isNodeLoginTestSuccess() {
            return isSuccess(ExtraStatusInfoType.NODE_LOGIN);
        }
        @Nullable
        public String getExtraInfo(@Nonnull String key) {
            return extraInfo.get(key);
        }

        @Nullable public List<String> getHardwareTypeNames() {
            return this.hardwareTypeNames;
        }

        public boolean hasHardwareType(String hardwareTypeName) {
            boolean res;
            if (hardwareTypeNames == null)
                res = false;
            else
                res = hardwareTypeNames.contains(hardwareTypeName);
//            LOG.debug("hasHardwareType() node="+name+" res="+res+" hardwareTypeNames="+hardwareTypeNames);
            return res;
        }

        @Nullable public List<String> getSliverTypeNames() {
            return this.sliverTypeNames;
        }

        public boolean hasSliverType(String sliverTypeName) {
            boolean res;
            if (sliverTypeNames == null)
                res = false;
            else
                res = sliverTypeNames.contains(sliverTypeName);
//            LOG.debug("hasSliverType() node="+name+" res="+res+" sliverTypeNames="+sliverTypeNames);
            return res;
        }

        @Override
        public int compareTo(@Nonnull Object o) {
            if (!(o instanceof ComponentInfo))
                throw new RuntimeException("Can't compare ComponentInfo to "+o.getClass().getName());

            ComponentInfo ci = (ComponentInfo) o;
            String s1 = this.getName() == null ? this.getUrn().toString() : this.getName();
            String s2 = ci.getName() == null ? ci.getUrn().toString() : ci.getName();

            if (s1 == null || s2 == null) {
                throw new NullPointerException("ComponentInfo must always have some ID");
            }
            return s1.compareTo(s2);
        }


        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            ComponentInfo that = (ComponentInfo) o;

            if (name != null ? !name.equals(that.name) : that.name != null) return false;
            if (urn != null ? !urn.equals(that.urn) : that.urn != null) return false;

            return true;
        }

        @Override
        public int hashCode() {
            int result = name != null ? name.hashCode() : 0;
            result = 31 * result + (urn != null ? urn.hashCode() : 0);
            return result;
        }

        @Override
        public String toString() {
            if (name != null)
                return name;
            return urn.getValue();
        }

        public NodeLocation getLocation() {
            return nodeLocation;
        }

        public boolean hasLocation() {
            return nodeLocation != null
                    && nodeLocation.getLatitude() != null
                    && nodeLocation.getLongitude() != null;
        }
    }
    /* something like:
    String colorName = server.getTestbed() == null ? null : server.getTestbed().getColor();

                if (colorName == null) colorName = "transparent";
                try {
                    color = Color.valueOf(colorName.toUpperCase());
                } catch (IllegalArgumentException ex) {
                    LOG.warn("Cannot convert color {}", testbedDesc.getString(TESTBED_COLOR, "transparent"), ex);
                }
    * */


    //provide info from which advertisement info such as sliver types is then derived
    void setAdvertisementRspec(Integer serverId, boolean availableAdvertisementRspec, ModelRspec modelRspec, Date fetchDate);
    void setServerGlimpse(Integer serverId, ServerGlimpse serverGlimpse);
    void setFedmonNodeTestInfo(Integer serverId, Map<GeniUrn, Result> fedmonResults);
    void initResourceClassAdvertisementInfo();

    @Nonnull ComponentInfo addUserComponentInfo(Integer serverId, GeniUrn componentId);
    @Nonnull HardwareTypeInfo addUserHardwareTypeInfo(Integer serverId, String hardwareTypeName, Boolean available);
    void clearUserInfo(Integer serverId);
}

